﻿local api = require "eu07.events"

-- tabela eventow ktore maja byc wywolane, jej celem jest mozliwosc ich "odwolania" w dowolnym momencie
-- eventQueue: {event, event, ...}
-- event: {name = nazwaFunkcjiWLua, params = {parametr, parametr, ...}, delay = opoznienie, conditions = {zmienna1 = wartosc1, zmienna2 = wartosc2, ...}}
eventQueue = {}
-- zliczanie "cykniec" zegara
ticks = 0
-- dokladnosc zegara (mniej = wieksza dokladnosc, wiekszy spam w logu i mniejsze FPS, wiecej = mniejsza dokladnosc, mniej spamu w logu i mniejsze obciazenie)
-- jednostka: sekundy / cykniecie
tickSize = 1

api.event_create("onstart_lua", 0, 0, function(event, activator)
	-- zegar - najwazniejsza jednostka calej symulacji, sluzy do eventow
	api.event_create("update", 0, 0, function(event, activator)
		api.event_dispatch_n("update", nil, tickSize)
		eventQueue_update(tickSize)
		ticks = ticks + 1
		-- losowanie uszkodzenia semafora
		-- bierzemy poprawke na dokladnosc zegara
		for name, data in pairs(devicesList["turn"]) do
			if data["broken"] then
				repairDevice("turn", name, repairChance / tickSize)
			else
				breakDevice("turn", name, breakChance / tickSize)
			end
		end
		for name, data in pairs(devicesList["sem"]) do
			if data["broken"] then
				repairDevice("sem", name, repairChance / tickSize)
			else
				breakDevice("sem", name, breakChance / tickSize)
			end
		end
	end)
	api.event_dispatch_n("update", nil, 0)
	-- losowanie zepsucia na starcie
	for name, data in pairs(devicesList["turn"]) do
		breakDevice("turn", name, startBreakChance)
	end
	for name, data in pairs(devicesList["sem"]) do
		breakDevice("sem", name, startBreakChance)
	end
	addLog(toString(executeFunc({["name"]="retrieveFromTable", ["params"]={"devicesList", {"sem", "stacja1_a", "broken"}}})))
end)

function linkEvent(name, commands)
	api.event_create(name, 0, 0, function(event, activator)
		for i = 1, #commands do
			local command = commands[i]
			executeFunc(command)
		end
	end)
	addLog("[Event] Utworzono event " .. name)
	addLog(toString(commands))
end

function eventQueue_update(delta)
	local i = 1
	while i <= #eventQueue do
		local event = eventQueue[i]
		-- Zmniejsz opoznienie o delte (ale tylko wtedy gdy warunki sa spelnione)
		if checkVariables(event["conditions"]) then
			event["delay"] = event["delay"] - delta
		end
		-- Jesli cale opoznienie zostalo juz wyczerpane (nadszedl czas wykonania polecenia) to wywolujemy ten event
		if event["delay"] <= 0 then
			eventQueue_execute(i)
			i = i - 1
		end
		i = i + 1
	end
end

function eventQueue_execute(id)
	-- Zapisujemy event ktory wykonujemy do zmiennej
	local event = eventQueue[id]
	-- Usuwamy event z kolejki
	table.remove(eventQueue, id)
	-- Wykonujemy funkcje w Lua
	executeFunc(event)
end

function eventQueue_add(data)
	table.insert(eventQueue, data)
	addLog("[Event] Dodano nowy event do kolejki...")
	addLog(toString(data))
end

function eventQueue_delete(name, params)
	-- oba parametry moga nie istniec (jesli oba nie beda istnialy to de facto usuniemy wszystko z kolejki ;] )
	addLog("[Event] Kasowanie eventu:")
	addLog("[Event] Nazwa: " .. name)
	addLog("[Event] Parametry: " .. toString(params))
	local i = 1
	while i <= #eventQueue do
		-- Sprawdz czy warunki pasuja i jesli tak, to usun z kolejki
		if eventQueue_check(i, name, params) then
			table.remove(eventQueue, i)
			addLog("[Event] Skasowano event numer " .. tostring(i) .. "!")
			i = i - 1
		end
		i = i + 1
	end
end

function eventQueue_check(id, name, params)
	-- eventQueue_check(idEventuKtorySprawdzamy, nazwaEventu (nil-able), {{param = parametr1, id = indeksParametru1 (nil-able)}, {param = parametr2, id = indeksParametru2 (nil-able)}, ...})
	-- co to jest indeksParametru?
	-- jesli mamy go podanego to szukamy <indeksParametru>-ty element i to on musi sie zgadzac, jesli go nie ma to zaliczamy bez wzgledu gdzie on jest
	-- tradycyjnie wyprowadzamy dane eventu ktore sprawdzamy
	local event = eventQueue[id]
	-- jesli mamy nazwe i nie pasuje, do widzenia
	if name and name ~= event["name"] then return false end
	-- jednak jesli sie upieklo to idziemy dalej, sprawdzamy liste parametrow (jesli istnieje!)
	if params then
		for i = 1, #params do
			-- wyciagamy parametr ktory sprawdzamy do zmiennej
			local param = params[i]
			-- dwa scenariusze, pierwszy: jesli mamy podany drugi parametr dla sprawdzenia parametrow - jego numer, to sprawdzamy ten konkretny parametr:
			if param["id"] and param["param"] ~= event["params"][param["id"]] then return false end
			-- drugi scenariusz, jesli nie mamy podanego konkretnego numeru parametru to iterujemy po wszystkich
			if not param["id"] and not tableHas(event["params"], param["param"]) then return false end
		end
	end
	return true
end