﻿local api = require "eu07.events"

-- szanse na zepsucie semafora/zwrotnicy na sekunde
breakChance = 10000000
-- szanse na zepsucie semafora/zwrotnicy na starcie
startBreakChance = 1000
-- szanse na naprawienie semafora/zwrotnicy na sekunde
repairChance = 1500

-- lista urzadzen, wszystkie dane sie tutaj znajduja
-- skladnia: {listaIzolacji, listaZwrotnic, listaSemaforow}
devicesList = {["track"]={}, ["turn"]={}, ["sem"]={}}
-- lista izolacji
-- skladnia: {nazwa = {occupied = zajetosc, reserved = rezerwacja, broken = awaria}, ...}
-- zajetosc i rezerwacja to boole
-- lista zwrotnic
-- skladnia: {nazwa = {state = stan, broken = awaria}, ...}
-- stan: 0 = plus (wprost), 1 = minus (w bok)
-- awaria to bool (w przyszlosci zostanie zamienione na int)
-- lista semaforow
-- skladnia: {nazwa = {state = stan, type = typ (true = ksztaltowy), prevSem = semaforyZalezne, repeaters = powtarzacze, routes = {przebieg, ...}, broken = awaria}, ...}
-- przebieg: nazwa = {speed = predkosc, nextSem = nastepnySemafor, turnouts = {zwrotnica, ...}, tracks = {izolacja, ...}}
-- zmienne stanu semaforow (0 - stoj, 1 - vmax, 2 - 40, 3 - 60, 4 - 100, 5 - Sz, 6 - Ms1, 7 - Ms2)

-- zwraca liste nazw urzadzen
function devicesTypeList(deviceType)
	return tableKeys(devicesList[deviceType])
end

-- inicjalizowanie wszystkiego
function init(data)
	-- struktura danych podana w pliku *.lua bezposrednio majacym kontakt ze sceneria
	for stationName, stationData in pairs(data) do
		-- SBL nie ma opoznienia ani szlak nie ma dyzurnego bo jest automatyczne, wiec wtedy nie robimy
		if stationData["type"] == 0 then
			iStacja(stationName)
		end
		local tracks = stationData["track"]
		for i = 1, #tracks do
			local track = objectName(tracks[i], stationData["shortName"])
			local trackData = {["occupied"]=false, ["reserved"]=false, ["broken"]=false}
			-- inicjalizowanie
			deviceInit("track", track, trackData)
		end
		local turnouts = stationData["turn"]
		for i = 1, #turnouts do
			local turnout = objectName(turnouts[i], stationData["shortName"])
			local turnoutData = {["state"]=0, ["broken"]=false}
			-- inicjalizowanie
			deviceInit("turn", turnout, turnoutData)
		end
		local semaphores = stationData["sem"]
		for semaphore, semaphoreData in pairs(semaphores) do
			semaphore = objectName(semaphore, stationData["shortName"])
			-- o ile w dwoch poprzednich przykladach musielismy stworzyc dane od zera, tu tylko uzupelniamy
			semaphoreData["state"] = 0
			semaphoreData["broken"] = false
			-- musimy troche nazw zamienic, bo podajemy skrotowo, ale w kodzie juz moga byc TYLKO pelne nazwy!
			semaphoreData["repeaters"] = objectName(semaphoreData["repeaters"], stationData["shortName"])
			for route, routeData in pairs(semaphoreData["routes"]) do
				for i = 1, #routeData do
					routeData[i]["nextSem"] = objectName(routeData[i]["nextSem"], stationData["shortName"])
					routeData[i]["prevSem"] = objectName(routeData[i]["prevSem"], stationData["shortName"])
					routeData[i]["turnouts"] = objectName(routeData[i]["turnouts"], stationData["shortName"])
					routeData[i]["tracks"] = objectName(routeData[i]["tracks"], stationData["shortName"])
				end
			end
			-- inicjalizowanie
			deviceInit("sem", semaphore, semaphoreData)
		end
	end
	addLog("dupa!!!")
	addLog(toString(devicesList["sem"]["stacja1_a"]))
end

function objectName(name, station)
	-- chodzi o to, ze jesli nazwa jest podana ze stacja, to zwraca ta nazwe z ta stacja, jesli nie ma podanej stacji, to ja dolacza
	if not name then return name end
	if type(name) == "string" then
		-- najpierw sprawdzamy czy mamy format z podkreslnikiem
		for i = 1, #name do
			if string.sub(name, i, i) == '_' then
				return name
			end
		end
		return station .. "_" .. name
	end
	-- hurtem:
	if type(name) == "table" then
		local names = {}
		for i = 1, #name do
			table.insert(names, objectName(name[i], station))
		end
		return names
	end
end

-- tory na stacjach (izolacje)
-- tworzymy dla kazdego toru eventy zmieniajace stan zmiennych
f_trackBusy = api.event_preparefunc(function(event, activator)
	local track = string.sub(api.event_getname(event), 1, -6)
	devicesList["track"][track]["occupied"] = true
end)
f_trackFree = api.event_preparefunc(function(event, activator)
	local track = string.sub(api.event_getname(event), 1, -6)
	devicesList["track"][track]["occupied"] = false
	devicesList["track"][track]["reserved"] = false
end)

-- semafory
-- tworzymy dla kazdego semafora polaczenia z eventami aby przy wygaszaniu semafora zauwazyly to zmienne i zmienily na 0
-- stoj na swietlnym
f_semStopS = api.event_preparefunc(function(event, activator)
	local semaphore = string.sub(api.event_getname(event), 1, -4)
	devicesList["sem"][semaphore]["state"] = 0
	-- ... oraz zeby wykasowaly wszystkie eventy ktore maja zaswiecic ten semafor (zapobiega to podaniu semafora za pociagiem)
	eventQueue_delete("setSemaphore", {{["param"]=semaphore, ["id"]=1}})
end)

-- stoj na ksztaltowym
f_semStopK = api.event_preparefunc(function(event, activator)
	local semaphore = string.sub(api.event_getname(event), 1, -5)
	devicesList["sem"][semaphore]["state"] = 0
	eventQueue_delete("setSemaphore", {{["param"]=semaphore, ["id"]=1}})
end)

-- Ms1
f_semStopT = api.event_preparefunc(function(event, activator)
	local semaphore = string.sub(api.event_getname(event), 1, -5)
	devicesList["sem"][semaphore]["state"] = 6
	eventQueue_delete("setSemaphore", {{["param"]=semaphore, ["id"]=1}})
end)

-- inicjalizacja urzadzenia
function deviceInit(deviceType, name, data)
	-- dodajemy urzadzenie do listy
	devicesList[deviceType][name] = data
	addLog(toString(data))
	-- tworzymy podlaczenia do eventow
	if deviceType == "track" then
		api.event_create(name .. ":busy", 0, 0, f_trackBusy)
		api.event_create(name .. ":free", 0, 0, f_trackFree)
		addLog("[Tory] Zainicjalizowano tor " .. name)
	end
	if deviceType == "turn" then
		addLog("[Tory] Zainicjalizowano zwrotnice " .. name)
	end
	if deviceType == "sem" then
		api.event_create(name .. "_s1", 0, 0, f_semStopS)
		api.event_create(name .. "_sr1", 0, 0, f_semStopK)
		api.event_create(name .. "_ms1", 0, 0, f_semStopT)
		addLog("[Tory] Zainicjalizowano semafor " .. name)
	end
end

-- stacje
-- funkcja inicjalizacyjna
function iStacja(stacja)
	-- pDyz = przymulenie dyzurnego, 0 = szybkosc swiatla, 60 = pijany jak ch*j
	_G["pDyz_" .. stacja] = api.random(0, 60)
	addLog("[Event] Zainicjalizowano stacje " .. stacja)
	addLog("[Event] Opoznienie stacji w sekundach (zaokraglone do 0.01): " .. tostring(round(_G["pDyz_" .. stacja], 2)))
end

function setRoute(semaphore, name, delay)
	-- Zwroc mozliwe przebiegi
	local routes = devicesList["sem"][semaphore]["routes"][name]
	-- Sprawdzamy kazdy przebieg w danym semaforze, czy pasuje nazwa
	for i = 1, #routes do
		local broken = false
		local route = routes[i]
		-- Sprawdzamy czy semafor w ogole da sie podac, jesli nie to koniec; odkladamy dopoki semafor nie zostanie naprawiony
		if devicesList["sem"][semaphore]["broken"] then
			eventQueue_add({["name"]="setRoute", ["params"]={semaphore, name, delay}, ["delay"]=60, ["conditions"]={{["name"]="retrieveFromTable", ["params"]={[{"devicesList", {"sem", semaphore, "broken"}}]=false}}}})
			return
		end
		-- Na poczatku musimy sprawdzic czy w ogole przebieg da sie ustawic, w tym celu sprawdzamy izolacje...
		local tracks = route["tracks"]
		for j = 1, #tracks do
			local track = tracks[j]
			if devicesList["track"][track]["occupied"] or devicesList["track"][track]["reserved"] then
				broken = true
				break
			end
		end
		-- oraz sprawdzamy czy zwrotnice na drodze przebiegu nie sa zepsute (jednak jesli zwrotnica jest ustawiona odpowiednio - zadziala).
		local turnouts = route["turnouts"]
		for j = 1, #turnouts do
			local turnout = turnouts[j]
			-- Zwrotnice w zaleznosciach to np. "zwr01+", sa one dwuczlonowe, cala nazwa z wyjatkiem ostatniego znaku to nazwa zwrotnicy, zapisujemy ja
			local name = string.sub(turnout, 1, -2)
			-- Reszte rowniez zapisujemy jako "+" (wprost) oraz "-" (w bok)
			local mark = string.sub(turnout, -1)
			-- Stan zwrotnicy musimy obliczyc bo polozenie zwrotnicy zapisywane jest jako 0 i 1, nie + i -, wiec konwertujemy
			if devicesList["turn"][name]["broken"] and turnoutState(mark) ~= devicesList["turn"][name]["state"] then
				broken = true
				break
			end
		end
		-- Jezeli wszystko jest OK, to wykonujemy przebieg
		if not broken then
			doRoute(semaphore, name, i, delay)
			return
		end
	end
	-- Jesli skonczyly sie przebiegi i zaden nie zostal wykonany, odkladamy wykonanie polecenia na pozniej (powtarzanie co minute)
	eventQueue_add({["name"]="setRoute", ["params"]={semaphore, name, delay}, ["delay"]=60, ["conditions"]={}})
end

function doRoute(semaphore, name, id, delay)
	-- Uzyskujemy dane o zadanym przebiegu
	local route = devicesList["sem"][semaphore]["routes"][name][id]
	-- Zrzucamy izolacje do zmiennej
	local tracks = route["tracks"]
	-- Rezerwujemy
	for i = 1, #tracks do
		devicesList["track"][tracks[i]]["reserved"] = true
	end
	-- Zrzucamy zwrotnice do zmiennej
	local turnouts = route["turnouts"]
	-- Musimy ustawic kazda zwrotnice
	for i = 1, #turnouts do
		-- Kazda ze zwrotnic zapisujemy do zmiennej
		local turnout = turnouts[i]
		-- Zwrotnice w zaleznosciach to np. "zwr01+", sa one dwuczlonowe, cala nazwa z wyjatkiem ostatniego znaku to nazwa zwrotnicy, zapisujemy ja
		local name = string.sub(turnout, 1, -2)
		-- Reszte rowniez zapisujemy jako "+" (wprost) oraz "-" (w bok)
		local mark = string.sub(turnout, -1)
		-- i zaktualizowac jej stan.
		setTurnout(name, turnoutState(mark), delay)
	end
	-- Na koniec pozostaje nam zaswiecenie semafora tego i poprzedniego (jesli istnieje).
	local nextSemState = nil
	if route["nextSem"] then
		nextSemState = devicesList["sem"][route["nextSem"]]["state"]
	end
	setSemaphore(semaphore, route["speed"], nextSemState, delay)
	if route["prevSem"] then
		setSemaphore(route["prevSem"], devicesList["sem"][route["prevSem"]]["state"], route["speed"], delay + 0.5) -- dodajemy 0.5 zeby poprawnie wyswietlilo na koniec
	end
end
--[[

to fix:

jesli mamy taki schemat posterunku:

      |
      C
 ---B]|[-*D-
      | /
      |/
      *
      A

	  
	  jesli najpierw podamy wjazd na semaforze A ale nie podamy wyjazdu na C (A = S5), a potem zaswiecimy semafor D (S2), to semafor A zmieni sygnal na S2 mimo sygnalu S1 na semaforze C
	  //fixed = dane o poprzednim semaforze w przebiegu
]]

function genOpoznienie(stacja)
	-- sWyk = szybkosc wykonania polecenia
	local sWyk = api.random(0, 15)
	return sWyk + _G["pDyz_" .. stacja]
end

function breakDevice(deviceType, name, chance)
	local rnd = api.random(0, chance)
	if rnd <= 1 then
		if deviceType == "turn" then
			addLog("[Event] Zepsula sie zwrotnica " .. name .. "!")
		end
		if deviceType == "sem" then
			addLog("[Event] Zepsul sie semafor " .. name .. "!")
		end
		devicesList[deviceType][name]["broken"] = true
	end
end

function repairDevice(deviceType, name, chance)
	local rnd = api.random(0, chance)
	if rnd <= 1 then
		if deviceType == "turn" then
			addLog("[Event] Naprawiono zwrotnice " .. name .. "!")
		end
		if deviceType == "sem" then
			addLog("[Event] Naprawiono semafor " .. name .. "!")
		end
		devicesList[deviceType][name]["broken"] = false
	end
end

-- ZARZADZANIE

-- Zwrotnice

function setTurnout(name, state, delay)
	if delay then
		eventQueue_add({["name"]="setTurnout", ["params"]={name, state}, ["delay"]=delay, ["conditions"]={}})
		return
	end
	if devicesList["turn"][name]["broken"] then
		addLog("[Tory] ERROR! Proba przestawienia uszkodzonej zwrotnicy " .. name .. "!")
		return
	end
	devicesList["turn"][name]["state"] = state
	api.event_dispatch_n(name .. turnoutMark(state), activator, 0)
	addLog("[Tory] Przestawiono zwrotnice " .. name .. " w polozenie " .. tostring(state) .. " (" .. turnoutMark(state) .. ")")
end

function turnoutState(mark)
	if mark == "+" then return 0 end
	if mark == "-" then return 1 end
end

function turnoutMark(state)
	if state == 0 then return "+" end
	if state == 1 then return "-" end
end

-- Semafory

function setSemaphore(semaphore, thisSignal, nextSignal, delay)
	if delay then
		eventQueue_add({["name"]="setSemaphore", ["params"]={semaphore, thisSignal, nextSignal}, ["delay"]=delay, ["conditions"]={}})
		return
	end
	if not nextSignal then nextSignal = 1 end
	if devicesList["sem"][semaphore]["broken"] then
		thisSignal = 5
	end
	local signal = signalName(0, devicesList["sem"][semaphore]["type"], thisSignal, nextSignal)
	devicesList["sem"][semaphore]["state"] = thisSignal
	api.event_dispatch_n(semaphore .. "_" .. signal, activator, 0)
	addLog("[Tory] Zapalono sygnal " .. signal .. " na semaforze " .. semaphore)
	local repeaters = devicesList["sem"][semaphore]["repeaters"]
	if repeaters then
		for i = 1, #repeaters do
			local repeater = repeaters[i]
			local repeaterSignal = signalName(2, false, thisSignal)
			api.event_dispatch_n(repeater .. "_" .. repeaterSignal, activator, 0)
		end
	end
end

function signalName(semType, shape, thisSignal, nextSignal)
	-- typSemafora:
	-- 0 = semafor / tarcza manewrowa / semafor SBL
	-- 1 = tarcza ostrzegawcza (ksztaltowy: tarcza ostrzegawcza dwustawna)
	-- 2 = powtarzacz (ksztaltowy: tarcza ostrzegawcza trzystawna)
	-- UWAGA!!! To oraz Sp odnosza sie do nastepnego semafora, ale podajemy tenSygnal!!!
	if not nextSignal then nextSignal = 0 end
	local signal = "sz1"
	if not shape then
		if semType == 0 then
			if thisSignal == 0 then signal = "s1" end
			if thisSignal == 1 and (nextSignal == 0 or nextSignal > 4) then signal = "s5" end
			if thisSignal == 1 and nextSignal == 1 then signal = "s2" end
			if thisSignal == 1 and (nextSignal == 2 or nextSignal == 3) then signal = "s4" end
			if thisSignal == 1 and nextSignal == 4 then signal = "s3" end
			if thisSignal == 2 and (nextSignal == 0 or nextSignal > 4) then signal = "s13" end
			if thisSignal == 2 and nextSignal == 1 then signal = "s10" end
			if thisSignal == 2 and (nextSignal == 2 or nextSignal == 3) then signal = "s12" end
			if thisSignal == 2 and nextSignal == 4 then signal = "s11" end
			if thisSignal == 3 and (nextSignal == 0 or nextSignal > 4) then signal = "s13a" end
			if thisSignal == 3 and nextSignal == 1 then signal = "s10a" end
			if thisSignal == 3 and (nextSignal == 2 or nextSignal == 3) then signal = "s12a" end
			if thisSignal == 3 and nextSignal == 4 then signal = "s11a" end
			if thisSignal == 4 and (nextSignal == 0 or nextSignal > 4) then signal = "s9" end
			if thisSignal == 4 and nextSignal == 1 then signal = "s6" end
			if thisSignal == 4 and (nextSignal == 2 or nextSignal == 3) then signal = "s8" end
			if thisSignal == 4 and nextSignal == 4 then signal = "s7" end
			if thisSignal == 5 then signal = "sz1" end
			if thisSignal == 6 then signal = "ms1" end
			if thisSignal == 7 then signal = "ms2" end
		end
		if semType == 1 then
			signal = "os1"
			if thisSignal == 0 or thisSignal > 4 then signal = "os1" end
			if thisSignal == 1 then signal = "os2" end
			if thisSignal == 2 or thisSignal == 3 then signal = "os3" end
			if thisSignal == 4 then signal = "os4" end
		end
		if semType == 2 then
			signal = "sp1"
			if thisSignal == 0 or thisSignal > 4 then signal = "sp1" end
			if thisSignal == 1 then signal = "sp2" end
			if thisSignal == 2 or thisSignal == 3 then signal = "sp4" end
			if thisSignal == 4 then signal = "sp3" end
		end
	end
	if shape then
		if semType == 0 then
			if thisSignal == 0 then signal = "sr1" end
			if thisSignal == 1 then signal = "sr2" end
			if thisSignal == 2 then signal = "sr3" end
			if thisSignal == 5 then signal = "sz1" end
			if thisSignal == 6 then signal = "ms1" end
			if thisSignal == 7 then signal = "ms2" end
		end
		if semType == 1 then
			signal = "od1"
			if thisSignal == 0 or thisSignal > 4 then signal = "od1" end
			if thisSignal == 1 then signal = "od2" end
		end
		if semType == 2 then
			signal = "ot1"
			if thisSignal == 0 or thisSignal > 4 then signal = "ot1" end
			if thisSignal == 1 then signal = "ot2" end
			if thisSignal == 2 then signal = "ot3" end
		end
	end
	return signal
end
